@@ -11,10 +11,11 @@ use winter_crypto::Hasher;
 
 use crate::{
     directory::get_marker_version,
+    errors::HistoryTreeNodeError,
     errors::{AkdError, AzksError, DirectoryError},
     node_state::{hash_label, NodeLabel},
     proof_structs::{HistoryProof, LookupProof, MembershipProof, NonMembershipProof, UpdateProof},
-    storage::types::AkdKey,
+    storage::types::AkdLabel,
     Direction, ARITY,
 };
 
@@ -28,25 +29,23 @@ pub fn verify_membership<H: Hasher>(
         if final_hash == root_hash {
             return Ok(());
         } else {
-            return Err(AkdError::AzksErr(AzksError::MembershipProofDidNotVerify(
+            return Err(AkdError::AzksErr(AzksError::VerifyMembershipProof(
                 "Membership proof for root did not verify".to_string(),
             )));
         }
     }
     let mut final_hash = H::merge(&[proof.hash_val, hash_label::<H>(proof.label)]);
     for i in (0..proof.dirs.len()).rev() {
-        final_hash = build_and_hash_layer::<H>(
-            proof.sibling_hashes[i],
-            proof.dirs[i],
-            final_hash,
-            proof.parent_labels[i],
-        )?;
+        // pull the hashes out of the label/hash(es) grouping
+        let hashes = proof.siblings[i].iter().map(|n| n.hash).collect();
+        final_hash =
+            build_and_hash_layer::<H>(hashes, proof.dirs[i], final_hash, proof.parent_labels[i])?;
     }
 
     if final_hash == root_hash {
         Ok(())
     } else {
-        return Err(AkdError::AzksErr(AzksError::MembershipProofDidNotVerify(
+        return Err(AkdError::AzksErr(AzksError::VerifyMembershipProof(
             format!(
                 "Membership proof for label {:?} did not verify",
                 proof.label
@@ -62,32 +61,32 @@ pub fn verify_nonmembership<H: Hasher>(
 ) -> Result<bool, AkdError> {
     let mut verified = true;
     let mut lcp_hash = H::hash(&[]);
-    let mut lcp_real = proof.longest_prefix_children_labels[0];
+    let mut lcp_real = proof.longest_prefix_children[0].label;
     for i in 0..ARITY {
         let child_hash = H::merge(&[
-            proof.longest_prefix_children_values[i],
-            hash_label::<H>(proof.longest_prefix_children_labels[i]),
+            proof.longest_prefix_children[i].hash,
+            hash_label::<H>(proof.longest_prefix_children[i].label),
         ]);
         lcp_hash = H::merge(&[lcp_hash, child_hash]);
-        lcp_real = lcp_real.get_longest_common_prefix(proof.longest_prefix_children_labels[i]);
+        lcp_real = lcp_real.get_longest_common_prefix(proof.longest_prefix_children[i].label);
     }
     // lcp_hash = H::merge(&[lcp_hash, hash_label::<H>(proof.longest_prefix)]);
     verified = verified && (lcp_hash == proof.longest_prefix_membership_proof.hash_val);
     if !verified {
-        return Err(AkdError::DirectoryErr(
-            DirectoryError::LookupVerificationErr("lcp_hash != longest_prefix_hash".to_string()),
-        ));
+        return Err(AkdError::Directory(DirectoryError::VerifyLookupProof(
+            "lcp_hash != longest_prefix_hash".to_string(),
+        )));
     }
-    let _sib_len = proof.longest_prefix_membership_proof.sibling_hashes.len();
+    let _sib_len = proof.longest_prefix_membership_proof.siblings.len();
     let _longest_prefix_verified =
         verify_membership(root_hash, &proof.longest_prefix_membership_proof)?;
     // The audit must have checked that this node is indeed the lcp of its children.
     // So we can just check that one of the children's lcp is = the proof.longest_prefix
     verified = verified && (proof.longest_prefix == lcp_real);
     if !verified {
-        return Err(AkdError::DirectoryErr(
-            DirectoryError::LookupVerificationErr("longest_prefix != lcp".to_string()),
-        ));
+        return Err(AkdError::Directory(DirectoryError::VerifyLookupProof(
+            "longest_prefix != lcp".to_string(),
+        )));
     }
     Ok(verified)
 }
@@ -95,7 +94,7 @@ pub fn verify_nonmembership<H: Hasher>(
 /// Verifies a lookup with respect to the root_hash
 pub fn lookup_verify<H: Hasher>(
     root_hash: H::Digest,
-    _akd_key: AkdKey,
+    _akd_key: AkdLabel,
     proof: LookupProof<H>,
 ) -> Result<(), AkdError> {
     let _epoch = proof.epoch;
@@ -112,7 +111,7 @@ pub fn lookup_verify<H: Hasher>(
     let existence_label = SeemlessDirectory::<S, H>::get_nodelabel(&uname, false, version);
     if existence_label != existence_proof.label {
         return Err(AkdError::DirectoryErr(
-            DirectoryError::LookupVerificationErr(
+            DirectoryError::VerifyLookupProof(
                 "Existence proof label does not match computed label".to_string(),
             ),
         ));
@@ -120,7 +119,7 @@ pub fn lookup_verify<H: Hasher>(
     let non_existence_label = SeemlessDirectory::<S, H>::get_nodelabel(&uname, true, version);
     if non_existence_label != freshness_proof.label {
         return Err(AkdError::DirectoryErr(
-            DirectoryError::LookupVerificationErr(
+            DirectoryError::VerifyLookupProof(
                 "Freshness proof label does not match computed label".to_string(),
             ),
         ));
@@ -128,7 +127,7 @@ pub fn lookup_verify<H: Hasher>(
     let marker_label = SeemlessDirectory::<S, H>::get_nodelabel(&uname, false, marker_version);
     if marker_label != marker_proof.label {
         return Err(AkdError::DirectoryErr(
-            DirectoryError::LookupVerificationErr(
+            DirectoryError::VerifyLookupProof(
                 "Marker proof label does not match computed label".to_string(),
             ),
         ));
@@ -146,7 +145,7 @@ pub fn lookup_verify<H: Hasher>(
 pub fn key_history_verify<H: Hasher>(
     root_hashes: Vec<H::Digest>,
     previous_root_hashes: Vec<Option<H::Digest>>,
-    uname: AkdKey,
+    uname: AkdLabel,
     proof: HistoryProof<H>,
 ) -> Result<(), AkdError> {
     for (count, update_proof) in proof.proofs.into_iter().enumerate() {
@@ -162,7 +161,7 @@ fn verify_single_update_proof<H: Hasher>(
     root_hash: H::Digest,
     previous_root_hash: Option<H::Digest>,
     proof: UpdateProof<H>,
-    uname: &AkdKey,
+    uname: &AkdLabel,
 ) -> Result<(), AkdError> {
     let epoch = proof.epoch;
     let _plaintext_value = &proof.plaintext_value;
@@ -177,13 +176,13 @@ fn verify_single_update_proof<H: Hasher>(
     // Need to include vrf verification
     // if label_at_ep != existence_at_ep_label {
     //     return Err(AkdError::DirectoryErr(
-    //         DirectoryError::KeyHistoryVerificationErr(
+    //         DirectoryError::VerifyKeyHistoryProof(
     //             format!("Label of user {:?}'s version {:?} at epoch {:?} does not match the one in the proof",
     //             uname, version, epoch))));
     // }
     verify_membership(root_hash, existence_at_ep)?;
     //     return Err(AkdError::DirectoryErr(
-    //         DirectoryError::KeyHistoryVerificationErr(format!(
+    //         DirectoryError::VerifyKeyHistoryProof(format!(
     //             "Existence proof of user {:?}'s version {:?} at epoch {:?} does not verify",
     //             uname, version, epoch
     //         )),
@@ -198,8 +197,7 @@ fn verify_single_update_proof<H: Hasher>(
             (version - 1),
             epoch
         );
-        let previous_null_err =
-            AkdError::DirectoryErr(DirectoryError::KeyHistoryVerificationErr(err_str));
+        let previous_null_err = AkdError::Directory(DirectoryError::VerifyKeyHistoryProof(err_str));
         let previous_val_stale_at_ep =
             previous_val_stale_at_ep.as_ref().ok_or(previous_null_err)?;
         verify_membership(root_hash, previous_val_stale_at_ep)?;
@@ -209,7 +207,7 @@ fn verify_single_update_proof<H: Hasher>(
         let root_hash = previous_root_hash.ok_or(AkdError::NoEpochGiven)?;
         verify_nonmembership(
             root_hash,
-            non_existence_before_ep.as_ref().ok_or_else(|| AkdError::DirectoryErr(DirectoryError::KeyHistoryVerificationErr(format!(
+            non_existence_before_ep.as_ref().ok_or_else(|| AkdError::Directory(DirectoryError::VerifyKeyHistoryProof(format!(
                 "Non-existence before this epoch proof of user {:?}'s version {:?} at epoch {:?} is None",
                 uname,
                 version,
@@ -223,8 +221,8 @@ fn verify_single_update_proof<H: Hasher>(
     for (i, ver) in (version + 1..(1 << next_marker)).enumerate() {
         let pf = &proof.non_existence_of_next_few[i];
         if !verify_nonmembership(root_hash, pf)? {
-            return Err(AkdError::DirectoryErr(
-                DirectoryError::KeyHistoryVerificationErr(
+            return Err(AkdError::Directory(
+                DirectoryError::VerifyKeyHistoryProof(
                     format!("Non-existence before epoch proof of user {:?}'s version {:?} at epoch {:?} does not verify",
                     uname, ver, epoch-1))));
         }
@@ -234,8 +232,8 @@ fn verify_single_update_proof<H: Hasher>(
         let ver = 1 << pow;
         let pf = &proof.non_existence_of_future_markers[i];
         if !verify_nonmembership(root_hash, pf)? {
-            return Err(AkdError::DirectoryErr(
-                DirectoryError::KeyHistoryVerificationErr(
+            return Err(AkdError::Directory(
+                DirectoryError::VerifyKeyHistoryProof(
                     format!("Non-existence before epoch proof of user {:?}'s version {:?} at epoch {:?} does not verify",
                     uname, ver, epoch-1))));
         }
@@ -246,15 +244,20 @@ fn verify_single_update_proof<H: Hasher>(
 
 /// Hashes all the children of a node, as well as their labels
 fn build_and_hash_layer<H: Hasher>(
-    hashes: [H::Digest; ARITY - 1],
+    hashes: Vec<H::Digest>,
     dir: Direction,
     ancestor_hash: H::Digest,
     parent_label: NodeLabel,
 ) -> Result<H::Digest, AkdError> {
-    let direction = dir.ok_or(AkdError::NoDirectionError)?;
-    let mut hashes_as_vec = hashes.to_vec();
-    hashes_as_vec.insert(direction, ancestor_hash);
-    Ok(hash_layer::<H>(hashes_as_vec, parent_label))
+    let direction = dir.ok_or_else(|| {
+        AkdError::HistoryTreeNode(HistoryTreeNodeError::NoDirection(
+            parent_label.get_val(),
+            None,
+        ))
+    })?;
+    let mut hashes_mut = hashes.to_vec();
+    hashes_mut.insert(direction, ancestor_hash);
+    Ok(hash_layer::<H>(hashes_mut, parent_label))
 }
 
 /// Helper for build_and_hash_layer
