@@ -9,12 +9,12 @@
 
 use crate::append_only_zks::Azks;
 
-use crate::node_state::NodeLabel;
+use crate::node_state::{Node, NodeLabel};
 use crate::proof_structs::*;
 
 use crate::errors::{AkdError, DirectoryError, HistoryTreeNodeError, StorageError};
 
-use crate::storage::types::{AkdKey, DbRecord, ValueState, ValueStateRetrievalFlag, Values};
+use crate::storage::types::{AkdLabel, AkdValue, DbRecord, ValueState, ValueStateRetrievalFlag};
 use crate::storage::Storage;
 
 use log::{debug, error, info};
@@ -29,14 +29,14 @@ use winter_crypto::Hasher;
 #[derive(Debug, Clone, Hash, Eq, PartialEq)]
 pub struct EpochHash<H: Hasher>(pub u64, pub H::Digest);
 
-impl Values {
+impl AkdValue {
     /// Gets a random value for a AKD
     pub fn random<R: RngCore + CryptoRng>(rng: &mut R) -> Self {
         Self(get_random_str(rng))
     }
 }
 
-impl AkdKey {
+impl AkdLabel {
     /// Creates a random key for a AKD
     pub fn random<R: RngCore + CryptoRng>(rng: &mut R) -> Self {
         Self(get_random_str(rng))
@@ -75,20 +75,20 @@ impl<S: Storage + Sync + Send> Directory<S> {
     /// Updates the directory to include the updated key-value pairs.
     pub async fn publish<H: Hasher>(
         &mut self,
-        updates: Vec<(AkdKey, Values)>,
+        updates: Vec<(AkdLabel, AkdValue)>,
         use_transaction: bool,
     ) -> Result<EpochHash<H>, AkdError> {
-        let mut update_set = Vec::<(NodeLabel, H::Digest)>::new();
+        let mut update_set = Vec::<Node<H>>::new();
         let mut user_data_update_set = Vec::<ValueState>::new();
         let next_epoch = self.current_epoch + 1;
 
-        let mut keys: Vec<AkdKey> = updates.iter().map(|(uname, _val)| uname.clone()).collect();
+        let mut keys: Vec<AkdLabel> = updates.iter().map(|(uname, _val)| uname.clone()).collect();
         // sort the keys, as inserting in primary-key order is more efficient for MySQL
         keys.sort_by(|a, b| a.0.cmp(&b.0));
 
         // we're only using the maximum "version" of the user's state at the last epoch
         // they were seen in the directory. Therefore we've minimized the call to only
-        // return a hashmap of AkdKey => u64 and not retrieving the other data which is not
+        // return a hashmap of AkdLabel => u64 and not retrieving the other data which is not
         // read (i.e. the actual _data_ payload).
         let all_user_versions_retrieved = self
             .storage
@@ -110,7 +110,10 @@ impl<S: Storage + Sync + Send> Directory<S> {
                     // Currently there's no blinding factor for the commitment.
                     // We'd want to change this later.
                     let value_to_add = H::hash(&Self::value_to_bytes(&val));
-                    update_set.push((label, value_to_add));
+                    update_set.push(Node::<H> {
+                        label,
+                        hash: value_to_add,
+                    });
                     let latest_state =
                         ValueState::new(uname, val, latest_version, label, next_epoch);
                     user_data_update_set.push(latest_state);
@@ -122,16 +125,21 @@ impl<S: Storage + Sync + Send> Directory<S> {
                     let fresh_label = Self::get_nodelabel::<H>(&uname, false, latest_version);
                     let stale_value_to_add = H::hash(&[0u8]);
                     let fresh_value_to_add = H::hash(&Self::value_to_bytes(&val));
-                    update_set.push((stale_label, stale_value_to_add));
-                    update_set.push((fresh_label, fresh_value_to_add));
+                    update_set.push(Node::<H> {
+                        label: stale_label,
+                        hash: stale_value_to_add,
+                    });
+                    update_set.push(Node::<H> {
+                        label: fresh_label,
+                        hash: fresh_value_to_add,
+                    });
                     let new_state =
                         ValueState::new(uname, val, latest_version, fresh_label, next_epoch);
                     user_data_update_set.push(new_state);
                 }
             }
         }
-        let insertion_set: Vec<(NodeLabel, H::Digest)> =
-            update_set.iter().map(|(x, y)| (*x, *y)).collect();
+        let insertion_set: Vec<Node<H>> = update_set.iter().copied().collect();
         // ideally the azks and the state would be updated together.
         // It may also make sense to have a temp version of the server's database
         let mut current_azks = self.retrieve_current_azks().await?;
@@ -139,11 +147,9 @@ impl<S: Storage + Sync + Send> Directory<S> {
         if use_transaction {
             if let false = self.storage.begin_transaction().await {
                 error!("Transaction is already active");
-                return Err(AkdError::HistoryTreeNodeErr(
-                    HistoryTreeNodeError::StorageError(StorageError::SetError(
-                        "Transaction is already active".to_string(),
-                    )),
-                ));
+                return Err(AkdError::HistoryTreeNode(HistoryTreeNodeError::Storage(
+                    StorageError::SetData("Transaction is already active".to_string()),
+                )));
             }
         }
         info!("Starting database insertion");
@@ -163,9 +169,9 @@ impl<S: Storage + Sync + Send> Directory<S> {
             if let Err(err) = self.storage.commit_transaction().await {
                 // ignore any rollback error(s)
                 let _ = self.storage.rollback_transaction().await;
-                return Err(AkdError::HistoryTreeNodeErr(
-                    HistoryTreeNodeError::StorageError(err),
-                ));
+                return Err(AkdError::HistoryTreeNode(HistoryTreeNodeError::Storage(
+                    err,
+                )));
             } else {
                 debug!("Transaction committed");
             }
@@ -183,7 +189,7 @@ impl<S: Storage + Sync + Send> Directory<S> {
     }
 
     /// Provides proof for correctness of latest version
-    pub async fn lookup<H: Hasher>(&self, uname: AkdKey) -> Result<LookupProof<H>, AkdError> {
+    pub async fn lookup<H: Hasher>(&self, uname: AkdLabel) -> Result<LookupProof<H>, AkdError> {
         match self
             .storage
             .get_user_state(&uname, ValueStateRetrievalFlag::MaxEpoch)
@@ -191,9 +197,10 @@ impl<S: Storage + Sync + Send> Directory<S> {
         {
             Err(_) => {
                 // Need to throw an error
-                Err(AkdError::DirectoryErr(
-                    DirectoryError::LookedUpNonExistentUser(uname.0, self.current_epoch),
-                ))
+                Err(AkdError::Directory(DirectoryError::NonExistentUser(
+                    uname.0,
+                    self.current_epoch,
+                )))
             }
             Ok(latest_st) => {
                 // Need to account for the case where the latest state is
@@ -233,7 +240,7 @@ impl<S: Storage + Sync + Send> Directory<S> {
     /// It also returns the proof of the latest version being served at all times.
     pub async fn key_history<H: Hasher>(
         &self,
-        uname: &AkdKey,
+        uname: &AkdLabel,
     ) -> Result<HistoryProof<H>, AkdError> {
         let username = uname.0.to_string();
         if let Ok(this_user_data) = self.storage.get_user_data(uname).await {
@@ -244,9 +251,10 @@ impl<S: Storage + Sync + Send> Directory<S> {
             }
             Ok(HistoryProof { proofs })
         } else {
-            Err(AkdError::DirectoryErr(
-                DirectoryError::LookedUpNonExistentUser(username, self.current_epoch),
-            ))
+            Err(AkdError::Directory(DirectoryError::NonExistentUser(
+                username,
+                self.current_epoch,
+            )))
         }
     }
 
@@ -264,11 +272,11 @@ impl<S: Storage + Sync + Send> Directory<S> {
     }
 
     /// Retrieves the current azks
-    pub async fn retrieve_current_azks(&self) -> Result<Azks, crate::errors::StorageError> {
+    pub async fn retrieve_current_azks(&self) -> Result<Azks, crate::errors::AkdError> {
         Directory::get_azks_from_storage(&self.storage).await
     }
 
-    async fn get_azks_from_storage(storage: &S) -> Result<Azks, crate::errors::StorageError> {
+    async fn get_azks_from_storage(storage: &S) -> Result<Azks, crate::errors::AkdError> {
         let got = storage
             .get::<Azks>(crate::append_only_zks::DEFAULT_AZKS_KEY)
             .await?;
@@ -276,8 +284,8 @@ impl<S: Storage + Sync + Send> Directory<S> {
             DbRecord::Azks(azks) => Ok(azks),
             _ => {
                 error!("No AZKS can be found. You should re-initialize the directory to create a new one");
-                Err(crate::errors::StorageError::GetError(String::from(
-                    "Not found",
+                Err(crate::errors::AkdError::AzksNotFound(String::from(
+                    "AZKS not found in storage.",
                 )))
             }
         }
@@ -286,7 +294,7 @@ impl<S: Storage + Sync + Send> Directory<S> {
     /// HELPERS ///
 
     #[allow(unused)]
-    fn username_to_nodelabel(_uname: &AkdKey) -> NodeLabel {
+    fn username_to_nodelabel(_uname: &AkdLabel) -> NodeLabel {
         // this function will need to read the VRF key off some function
         unimplemented!()
     }
@@ -296,7 +304,11 @@ impl<S: Storage + Sync + Send> Directory<S> {
     /// Returns the tree nodelabel that corresponds to a version of the akdkey argument.
     /// The stale boolean here is to indicate whether we are getting the nodelabel for a fresh version,
     /// or a version that we are retiring.
-    pub(crate) fn get_nodelabel<H: Hasher>(uname: &AkdKey, stale: bool, version: u64) -> NodeLabel {
+    pub(crate) fn get_nodelabel<H: Hasher>(
+        uname: &AkdLabel,
+        stale: bool,
+        version: u64,
+    ) -> NodeLabel {
         // this function will need to read the VRF key using some function
         let name_hash_bytes = H::hash(uname.0.as_bytes());
         let mut stale_bytes = &[1u8];
@@ -315,14 +327,14 @@ impl<S: Storage + Sync + Send> Directory<S> {
 
     // FIXME: Make a real commitment here, alongwith a blinding factor.
     /// Gets the bytes for a value.
-    pub fn value_to_bytes(_value: &Values) -> [u8; 64] {
+    pub fn value_to_bytes(_value: &AkdValue) -> [u8; 64] {
         [0u8; 64]
         // unimplemented!()
     }
 
     async fn create_single_update_proof<H: Hasher>(
         &self,
-        uname: &AkdKey,
+        uname: &AkdLabel,
         user_state: &ValueState,
     ) -> Result<UpdateProof<H>, AkdError> {
         let epoch = user_state.epoch;
@@ -498,7 +510,7 @@ mod tests {
         let mut akd = Directory::<_>::new::<Blake3>(&db).await?;
 
         akd.publish::<Blake3>(
-            vec![(AkdKey("hello".to_string()), Values("world".to_string()))],
+            vec![(AkdLabel("hello".to_string()), AkdValue("world".to_string()))],
             false,
         )
         .await?;
@@ -512,19 +524,22 @@ mod tests {
 
         akd.publish::<Blake3>(
             vec![
-                (AkdKey("hello".to_string()), Values("world".to_string())),
-                (AkdKey("hello2".to_string()), Values("world2".to_string())),
+                (AkdLabel("hello".to_string()), AkdValue("world".to_string())),
+                (
+                    AkdLabel("hello2".to_string()),
+                    AkdValue("world2".to_string()),
+                ),
             ],
             false,
         )
         .await?;
 
-        let lookup_proof = akd.lookup(AkdKey("hello".to_string())).await?;
+        let lookup_proof = akd.lookup(AkdLabel("hello".to_string())).await?;
         let current_azks = akd.retrieve_current_azks().await?;
         let root_hash = akd.get_root_hash::<Blake3>(&current_azks).await?;
         lookup_verify::<Blake3_256<BaseElement>>(
             root_hash,
-            AkdKey("hello".to_string()),
+            AkdLabel("hello".to_string()),
             lookup_proof,
         )?;
         Ok(())
@@ -537,8 +552,11 @@ mod tests {
 
         akd.publish::<Blake3>(
             vec![
-                (AkdKey("hello".to_string()), Values("world".to_string())),
-                (AkdKey("hello2".to_string()), Values("world2".to_string())),
+                (AkdLabel("hello".to_string()), AkdValue("world".to_string())),
+                (
+                    AkdLabel("hello2".to_string()),
+                    AkdValue("world2".to_string()),
+                ),
             ],
             false,
         )
@@ -546,8 +564,11 @@ mod tests {
 
         akd.publish::<Blake3>(
             vec![
-                (AkdKey("hello".to_string()), Values("world".to_string())),
-                (AkdKey("hello2".to_string()), Values("world2".to_string())),
+                (AkdLabel("hello".to_string()), AkdValue("world".to_string())),
+                (
+                    AkdLabel("hello2".to_string()),
+                    AkdValue("world2".to_string()),
+                ),
             ],
             false,
         )
@@ -555,8 +576,14 @@ mod tests {
 
         akd.publish::<Blake3>(
             vec![
-                (AkdKey("hello".to_string()), Values("world3".to_string())),
-                (AkdKey("hello2".to_string()), Values("world4".to_string())),
+                (
+                    AkdLabel("hello".to_string()),
+                    AkdValue("world3".to_string()),
+                ),
+                (
+                    AkdLabel("hello2".to_string()),
+                    AkdValue("world4".to_string()),
+                ),
             ],
             false,
         )
@@ -564,8 +591,14 @@ mod tests {
 
         akd.publish::<Blake3>(
             vec![
-                (AkdKey("hello3".to_string()), Values("world".to_string())),
-                (AkdKey("hello4".to_string()), Values("world2".to_string())),
+                (
+                    AkdLabel("hello3".to_string()),
+                    AkdValue("world".to_string()),
+                ),
+                (
+                    AkdLabel("hello4".to_string()),
+                    AkdValue("world2".to_string()),
+                ),
             ],
             false,
         )
@@ -573,8 +606,8 @@ mod tests {
 
         akd.publish::<Blake3>(
             vec![(
-                AkdKey("hello".to_string()),
-                Values("world_updated".to_string()),
+                AkdLabel("hello".to_string()),
+                AkdValue("world_updated".to_string()),
             )],
             false,
         )
@@ -582,50 +615,56 @@ mod tests {
 
         akd.publish::<Blake3>(
             vec![
-                (AkdKey("hello3".to_string()), Values("world6".to_string())),
-                (AkdKey("hello4".to_string()), Values("world12".to_string())),
+                (
+                    AkdLabel("hello3".to_string()),
+                    AkdValue("world6".to_string()),
+                ),
+                (
+                    AkdLabel("hello4".to_string()),
+                    AkdValue("world12".to_string()),
+                ),
             ],
             false,
         )
         .await?;
 
-        let history_proof = akd.key_history(&AkdKey("hello".to_string())).await?;
+        let history_proof = akd.key_history(&AkdLabel("hello".to_string())).await?;
         let (root_hashes, previous_root_hashes) =
             get_key_history_hashes(&akd, &history_proof).await?;
         key_history_verify::<Blake3_256<BaseElement>>(
             root_hashes,
             previous_root_hashes,
-            AkdKey("hello".to_string()),
+            AkdLabel("hello".to_string()),
             history_proof,
         )?;
 
-        let history_proof = akd.key_history(&AkdKey("hello2".to_string())).await?;
+        let history_proof = akd.key_history(&AkdLabel("hello2".to_string())).await?;
         let (root_hashes, previous_root_hashes) =
             get_key_history_hashes(&akd, &history_proof).await?;
         key_history_verify::<Blake3_256<BaseElement>>(
             root_hashes,
             previous_root_hashes,
-            AkdKey("hello2".to_string()),
+            AkdLabel("hello2".to_string()),
             history_proof,
         )?;
 
-        let history_proof = akd.key_history(&AkdKey("hello3".to_string())).await?;
+        let history_proof = akd.key_history(&AkdLabel("hello3".to_string())).await?;
         let (root_hashes, previous_root_hashes) =
             get_key_history_hashes(&akd, &history_proof).await?;
         key_history_verify::<Blake3_256<BaseElement>>(
             root_hashes,
             previous_root_hashes,
-            AkdKey("hello3".to_string()),
+            AkdLabel("hello3".to_string()),
             history_proof,
         )?;
 
-        let history_proof = akd.key_history(&AkdKey("hello4".to_string())).await?;
+        let history_proof = akd.key_history(&AkdLabel("hello4".to_string())).await?;
         let (root_hashes, previous_root_hashes) =
             get_key_history_hashes(&akd, &history_proof).await?;
         key_history_verify::<Blake3_256<BaseElement>>(
             root_hashes,
             previous_root_hashes,
-            AkdKey("hello4".to_string()),
+            AkdLabel("hello4".to_string()),
             history_proof,
         )?;
 
@@ -640,8 +679,11 @@ mod tests {
 
         akd.publish::<Blake3>(
             vec![
-                (AkdKey("hello".to_string()), Values("world".to_string())),
-                (AkdKey("hello2".to_string()), Values("world2".to_string())),
+                (AkdLabel("hello".to_string()), AkdValue("world".to_string())),
+                (
+                    AkdLabel("hello2".to_string()),
+                    AkdValue("world2".to_string()),
+                ),
             ],
             false,
         )
@@ -649,8 +691,11 @@ mod tests {
 
         akd.publish::<Blake3>(
             vec![
-                (AkdKey("hello".to_string()), Values("world".to_string())),
-                (AkdKey("hello2".to_string()), Values("world2".to_string())),
+                (AkdLabel("hello".to_string()), AkdValue("world".to_string())),
+                (
+                    AkdLabel("hello2".to_string()),
+                    AkdValue("world2".to_string()),
+                ),
             ],
             false,
         )
@@ -658,8 +703,14 @@ mod tests {
 
         akd.publish::<Blake3>(
             vec![
-                (AkdKey("hello".to_string()), Values("world3".to_string())),
-                (AkdKey("hello2".to_string()), Values("world4".to_string())),
+                (
+                    AkdLabel("hello".to_string()),
+                    AkdValue("world3".to_string()),
+                ),
+                (
+                    AkdLabel("hello2".to_string()),
+                    AkdValue("world4".to_string()),
+                ),
             ],
             false,
         )
@@ -667,8 +718,14 @@ mod tests {
 
         akd.publish::<Blake3>(
             vec![
-                (AkdKey("hello3".to_string()), Values("world".to_string())),
-                (AkdKey("hello4".to_string()), Values("world2".to_string())),
+                (
+                    AkdLabel("hello3".to_string()),
+                    AkdValue("world".to_string()),
+                ),
+                (
+                    AkdLabel("hello4".to_string()),
+                    AkdValue("world2".to_string()),
+                ),
             ],
             false,
         )
@@ -676,8 +733,8 @@ mod tests {
 
         akd.publish::<Blake3>(
             vec![(
-                AkdKey("hello".to_string()),
-                Values("world_updated".to_string()),
+                AkdLabel("hello".to_string()),
+                AkdValue("world_updated".to_string()),
             )],
             false,
         )
@@ -685,8 +742,14 @@ mod tests {
 
         akd.publish::<Blake3>(
             vec![
-                (AkdKey("hello3".to_string()), Values("world6".to_string())),
-                (AkdKey("hello4".to_string()), Values("world12".to_string())),
+                (
+                    AkdLabel("hello3".to_string()),
+                    AkdValue("world6".to_string()),
+                ),
+                (
+                    AkdLabel("hello4".to_string()),
+                    AkdValue("world12".to_string()),
+                ),
             ],
             false,
         )
