@@ -82,14 +82,9 @@ impl Storable for HistoryTreeNode {
 
     fn get_full_binary_key_id(key: &NodeKey) -> Vec<u8> {
         let mut result = vec![StorageType::HistoryTreeNode as u8];
-        let len_bytes = key.0.len.to_be_bytes();
-        for byte in &len_bytes {
-            result.push(*byte);
-        }
-        let parts = key.0.val.to_be_bytes();
-        for byte in &parts {
-            result.push(*byte);
-        }
+        result.extend_from_slice(&key.0.len.to_be_bytes());
+        result.extend_from_slice(&key.0.val.to_be_bytes());
+
         result
     }
 
@@ -145,7 +140,7 @@ impl HistoryTreeNode {
     ) -> Result<HistoryTreeNode, StorageError> {
         match storage.get::<HistoryTreeNode>(key).await? {
             DbRecord::HistoryTreeNode(node) => Ok(node),
-            _ => Err(StorageError::GetError(String::from("Not found"))),
+            _ => Err(StorageError::GetData(String::from("Not found"))),
         }
     }
 
@@ -159,7 +154,7 @@ impl HistoryTreeNode {
             if let DbRecord::HistoryTreeNode(node) = node {
                 nodes.push(node);
             } else {
-                return Err(StorageError::GetError(String::from(
+                return Err(StorageError::GetData(String::from(
                     "Batch retrieve returned types <> HistoryTreeNode",
                 )));
             }
@@ -302,7 +297,7 @@ impl HistoryTreeNode {
                 let child_st = self
                     .get_child_at_epoch::<_, H>(storage, self.get_latest_epoch()?, dir_leaf)
                     .await?
-                    .ok_or(HistoryTreeNodeError::NoChildInTreeAtEpoch(
+                    .ok_or(HistoryTreeNodeError::NoChildAtEpoch(
                         self.get_latest_epoch()?,
                         dir_leaf.unwrap(),
                     ))?;
@@ -412,13 +407,13 @@ impl HistoryTreeNode {
         match get_state_map(storage, parent, epoch).await {
             Err(_) => Err(HistoryTreeNodeError::ParentNextEpochInvalid(epoch)),
             Ok(parent_state) => match parent.get_direction_at_ep(storage, self, epoch).await? {
-                None => Err(HistoryTreeNodeError::HashUpdateOnlyAllowedAfterNodeInsertion),
+                None => Err(HistoryTreeNodeError::HashUpdateOrderInconsistent),
                 Some(s_dir) => {
                     let mut parent_updated_state = parent_state;
                     let mut self_child_state =
                         parent_updated_state
                             .get_child_state_in_dir(s_dir)
-                            .ok_or(HistoryTreeNodeError::NoChildInTreeAtEpoch(epoch, s_dir))?;
+                            .ok_or(HistoryTreeNodeError::NoChildAtEpoch(epoch, s_dir))?;
                     self_child_state.hash_val = from_digest::<H>(new_hash_val)?;
                     parent_updated_state.child_states[s_dir] = Some(self_child_state);
                     parent_updated_state.key = NodeStateKey(parent.label, epoch);
@@ -480,9 +475,9 @@ impl HistoryTreeNode {
         }
 
         let dir = direction.map_or(
-            Err(HistoryTreeNodeError::NoDirectionInSettingChild(
+            Err(HistoryTreeNodeError::NoDirection(
                 self.label.get_val(),
-                child_node.label.get_val(),
+                Some(child_node.label.get_val()),
             )),
             Ok,
         )?;
@@ -560,7 +555,7 @@ impl HistoryTreeNode {
         Ok(self
             .get_child_at_epoch::<_, H>(storage, epoch, dir)
             .await?
-            .ok_or_else(|| HistoryTreeNodeError::NoChildInTreeAtEpoch(epoch, dir.unwrap()))?
+            .ok_or_else(|| HistoryTreeNodeError::NoChildAtEpoch(epoch, dir.unwrap()))?
             .label)
     }
 
@@ -615,10 +610,13 @@ impl HistoryTreeNode {
         direction: Direction,
     ) -> Result<Option<HistoryChildState>, HistoryTreeNodeError> {
         match direction {
-            Direction::None => Err(HistoryTreeNodeError::DirectionIsNone),
+            Direction::None => Err(HistoryTreeNodeError::NoDirection(
+                self.label.get_val(),
+                None,
+            )),
             Direction::Some(dir) => {
                 if self.get_birth_epoch() > epoch {
-                    Err(HistoryTreeNodeError::NoChildInTreeAtEpoch(epoch, dir))
+                    Err(HistoryTreeNodeError::NoChildAtEpoch(epoch, dir))
                 } else {
                     let chosen_ep = {
                         if self.last_epoch <= epoch {
@@ -657,7 +655,10 @@ impl HistoryTreeNode {
         direction: Direction,
     ) -> Result<Option<HistoryChildState>, HistoryTreeNodeError> {
         match direction {
-            Direction::None => Err(HistoryTreeNodeError::DirectionIsNone),
+            Direction::None => Err(HistoryTreeNodeError::NoDirection(
+                self.label.get_val(),
+                None,
+            )),
             Direction::Some(dir) => Ok(get_state_map(storage, self, epoch)
                 .await
                 .map(|curr| curr.get_child_state_in_dir(dir))?),
@@ -670,7 +671,7 @@ impl HistoryTreeNode {
         epoch: u64,
     ) -> Result<HistoryNodeState, HistoryTreeNodeError> {
         if self.get_birth_epoch() > epoch {
-            Err(HistoryTreeNodeError::NodeDidNotExistAtEp(self.label, epoch))
+            Err(HistoryTreeNodeError::NonexistentAtEpoch(self.label, epoch))
         } else {
             let chosen_ep = {
                 if self.last_epoch <= epoch {
@@ -704,7 +705,7 @@ impl HistoryTreeNode {
     ) -> Result<HistoryNodeState, HistoryTreeNodeError> {
         get_state_map(storage, self, epoch)
             .await
-            .map_err(|_| HistoryTreeNodeError::NodeDidNotHaveExistingStateAtEp(self.label, epoch))
+            .map_err(|_| HistoryTreeNodeError::NoStateAtEpoch(self.label, epoch))
     }
 
     /* Functions for compression-related operations */
@@ -807,13 +808,12 @@ pub(crate) async fn get_leaf_node<H: Hasher, S: Storage + Sync + Send>(
 
 pub(crate) async fn get_leaf_node_without_hashing<H: Hasher, S: Storage + Sync + Send>(
     storage: &S,
-    label: NodeLabel,
-    value: H::Digest,
+    node: Node<H>,
     parent: NodeLabel,
     birth_epoch: u64,
 ) -> Result<HistoryTreeNode, HistoryTreeNodeError> {
-    let node = HistoryTreeNode {
-        label,
+    let history_node = HistoryTreeNode {
+        label: node.label,
         birth_epoch,
         last_epoch: birth_epoch,
         parent,
@@ -821,12 +821,12 @@ pub(crate) async fn get_leaf_node_without_hashing<H: Hasher, S: Storage + Sync +
     };
 
     let mut new_state: HistoryNodeState =
-        HistoryNodeState::new::<H>(NodeStateKey(node.label, birth_epoch));
-    new_state.value = from_digest::<H>(value)?;
+        HistoryNodeState::new::<H>(NodeStateKey(history_node.label, birth_epoch));
+    new_state.value = from_digest::<H>(node.hash)?;
 
     set_state_map(storage, new_state).await?;
 
-    Ok(node)
+    Ok(history_node)
 }
 
 pub(crate) async fn set_state_map<S: Storage + Sync + Send>(
@@ -847,7 +847,7 @@ pub(crate) async fn get_state_map<S: Storage + Sync + Send>(
     {
         Ok(state)
     } else {
-        Err(StorageError::GetError(String::from("Not found")))
+        Err(StorageError::GetData(String::from("Not found")))
     }
 }
 
