@@ -20,6 +20,28 @@ use winter_crypto::Hasher;
 
 use rand::{CryptoRng, RngCore};
 
+/// Represents a node's label & associated hash
+#[derive(Debug)]
+pub struct Node<H: Hasher> {
+    /// the label associated with the accompanying hash
+    pub label: NodeLabel,
+    /// the hash associated to this label
+    pub hash: H::Digest,
+}
+
+// can't use #derive because it doesn't bind correctly
+// #derive and generics are not friendly; might make Debug weird too ...
+// see also:
+// https://users.rust-lang.org/t/why-does-deriving-clone-not-work-in-this-case-but-implementing-manually-does/29075
+// https://github.com/rust-lang/rust/issues/26925
+impl<H: Hasher> Copy for Node<H> {}
+
+impl<H: Hasher> Clone for Node<H> {
+    fn clone(&self) -> Node<H> {
+        *self
+    }
+}
+
 /// The NodeLabel struct represents the label for a HistoryTreeNode.
 /// Since the label itself may have any number of zeros pre-pended,
 /// just using a native type, unless it is a bit-vector, wouldn't work.
@@ -210,17 +232,10 @@ impl Storable for HistoryNodeState {
 
     fn get_full_binary_key_id(key: &NodeStateKey) -> Vec<u8> {
         let mut result = vec![StorageType::HistoryNodeState as u8];
-        let len_bytes = key.0.len.to_be_bytes();
-        for byte in &len_bytes {
-            result.push(*byte);
-        }
+        result.extend_from_slice(&key.0.len.to_be_bytes());
+        result.extend_from_slice(&key.0.val.to_be_bytes());
+        result.extend_from_slice(&key.1.to_be_bytes());
 
-        let parts: [[u8; 8]; 2] = [key.0.val.to_be_bytes(), key.1.to_be_bytes()];
-        for iarray in &parts {
-            for byte in iarray {
-                result.push(*byte);
-            }
-        }
         result
     }
 
