@@ -8,7 +8,7 @@
 use crate::errors::StorageError;
 use crate::storage::transaction::Transaction;
 use crate::storage::types::{
-    AkdKey, DbRecord, KeyData, StorageType, ValueState, ValueStateKey, ValueStateRetrievalFlag,
+    AkdLabel, DbRecord, KeyData, StorageType, ValueState, ValueStateKey, ValueStateRetrievalFlag,
 };
 use crate::storage::{Storable, Storage};
 use async_trait::async_trait;
@@ -146,7 +146,7 @@ impl Storage for AsyncInMemoryDatabase {
                         return Ok(DbRecord::ValueState(item.clone()));
                     }
                 }
-                return Err(StorageError::GetError("Not found".to_string()));
+                return Err(StorageError::GetData("Not found".to_string()));
             }
         }
         // fallback to regular get/set db
@@ -154,7 +154,7 @@ impl Storage for AsyncInMemoryDatabase {
         if let Some(result) = (*guard).get(&bin_id).cloned() {
             Ok(result)
         } else {
-            Err(StorageError::GetError("Not found".to_string()))
+            Err(StorageError::GetData("Not found".to_string()))
         }
     }
 
@@ -174,7 +174,7 @@ impl Storage for AsyncInMemoryDatabase {
     }
 
     /// Retrieve the user data for a given user
-    async fn get_user_data(&self, username: &AkdKey) -> Result<KeyData, StorageError> {
+    async fn get_user_data(&self, username: &AkdLabel) -> Result<KeyData, StorageError> {
         let guard = self.user_info.read().await;
         if let Some(result) = guard.get(&username.0) {
             let mut results: Vec<ValueState> = result.to_vec();
@@ -183,14 +183,14 @@ impl Storage for AsyncInMemoryDatabase {
 
             Ok(KeyData { states: results })
         } else {
-            Err(StorageError::GetError("Not found".to_string()))
+            Err(StorageError::GetData("Not found".to_string()))
         }
     }
 
     /// Retrieve a specific state for a given user
     async fn get_user_state(
         &self,
-        username: &AkdKey,
+        username: &AkdLabel,
         flag: ValueStateRetrievalFlag,
     ) -> Result<ValueState, StorageError> {
         let intermediate = self.get_user_data(username).await?.states;
@@ -250,18 +250,18 @@ impl Storage for AsyncInMemoryDatabase {
                 }
             }
         }
-        Err(StorageError::GetError(String::from("Not found")))
+        Err(StorageError::GetData(String::from("Not found")))
     }
 
     async fn get_user_state_versions(
         &self,
-        keys: &[AkdKey],
+        keys: &[AkdLabel],
         flag: ValueStateRetrievalFlag,
-    ) -> Result<HashMap<AkdKey, u64>, StorageError> {
+    ) -> Result<HashMap<AkdLabel, u64>, StorageError> {
         let mut map = HashMap::new();
         for username in keys.iter() {
             if let Ok(result) = self.get_user_state(username, flag).await {
-                map.insert(AkdKey(result.username.0.clone()), result.version);
+                map.insert(AkdLabel(result.username.0.clone()), result.version);
             }
         }
         Ok(map)
@@ -298,7 +298,7 @@ impl Storage for AsyncInMemoryDatabase {
                 return Ok(item);
             }
         }
-        Err(StorageError::GetError(format!(
+        Err(StorageError::GetData(format!(
             "Node (val: {}, len: {}) did not exist <= epoch {}",
             node_label.val, node_label.len, epoch_in_question
         )))
@@ -513,7 +513,7 @@ impl Storage for AsyncInMemoryDbWithCache {
                         return Ok(DbRecord::ValueState(item.clone()));
                     }
                 }
-                return Err(StorageError::GetError("Not found".to_string()));
+                return Err(StorageError::GetData("Not found".to_string()));
             }
         }
         let mut stats = self.stats.write().await;
@@ -532,7 +532,7 @@ impl Storage for AsyncInMemoryDbWithCache {
 
                     Ok(result)
                 } else {
-                    Err(StorageError::GetError("Not found".to_string()))
+                    Err(StorageError::GetData("Not found".to_string()))
                 }
             }
         }
@@ -552,7 +552,7 @@ impl Storage for AsyncInMemoryDbWithCache {
         Ok(map)
     }
 
-    async fn get_user_data(&self, username: &AkdKey) -> Result<KeyData, StorageError> {
+    async fn get_user_data(&self, username: &AkdLabel) -> Result<KeyData, StorageError> {
         let guard = self.user_info.read().await;
         if let Some(result) = guard.get(&username.0) {
             let mut results: Vec<ValueState> = result.to_vec();
@@ -561,13 +561,13 @@ impl Storage for AsyncInMemoryDbWithCache {
 
             Ok(KeyData { states: results })
         } else {
-            Err(StorageError::GetError("Not found".to_string()))
+            Err(StorageError::GetData("Not found".to_string()))
         }
     }
 
     async fn get_user_state(
         &self,
-        username: &AkdKey,
+        username: &AkdLabel,
         flag: ValueStateRetrievalFlag,
     ) -> Result<ValueState, StorageError> {
         let intermediate = self.get_user_data(username).await?.states;
@@ -627,18 +627,18 @@ impl Storage for AsyncInMemoryDbWithCache {
                 }
             }
         }
-        Err(StorageError::GetError(String::from("Not found")))
+        Err(StorageError::GetData(String::from("Not found")))
     }
 
     async fn get_user_state_versions(
         &self,
-        keys: &[AkdKey],
+        keys: &[AkdLabel],
         flag: ValueStateRetrievalFlag,
-    ) -> Result<HashMap<AkdKey, u64>, StorageError> {
+    ) -> Result<HashMap<AkdLabel, u64>, StorageError> {
         let mut map = HashMap::new();
         for username in keys.iter() {
             if let Ok(result) = self.get_user_state(username, flag).await {
-                map.insert(AkdKey(result.username.0.clone()), result.version);
+                map.insert(AkdLabel(result.username.0.clone()), result.version);
             }
         }
         Ok(map)
@@ -649,7 +649,7 @@ impl Storage for AsyncInMemoryDbWithCache {
         node_label: crate::node_state::NodeLabel,
         epoch_in_question: u64,
     ) -> Result<u64, StorageError> {
-        let ids = (0..epoch_in_question)
+        let ids = (0..=epoch_in_question)
             .map(|epoch| crate::node_state::NodeStateKey(node_label, epoch))
             .collect::<Vec<_>>();
         let data = self
@@ -675,7 +675,7 @@ impl Storage for AsyncInMemoryDbWithCache {
                 return Ok(item);
             }
         }
-        Err(StorageError::GetError(format!(
+        Err(StorageError::GetData(format!(
             "Node (val: {}, len: {}) did not exist <= epoch {}",
             node_label.val, node_label.len, epoch_in_question
         )))
